/*
 *  This file if part of nzbget
 *
 *  Copyright (C) 2004  Sven Henkel <sidddy@users.sourceforge.net>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Revision: 1.22 $
 * $Date: 2005/07/12 09:37:48 $
 *
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <unistd.h>
#include <sys/time.h>

#include "ArticleDownloader.h"
#include "Decoder.h"
#include "global.h"
#include "GlobalControl.h"
#include "Frontend.h"

#include "MultiServerPool.h"

#include "NZBQueueCoordinator.h"

extern NZBQueueCoordinator* g_pNZBQueueCoordinator;
extern Frontend* g_pFrontend;
extern Options* g_pOptions;
extern float g_fDownloadRate;
extern bool g_bPause;

ArticleDownloader::ArticleDownloader(ArticleJob* pJob, char* szTempDir)
{
	m_pJob = pJob;
	m_pJob->SetStatus(JOBSTATUS_ASSIGNED);
	m_bCancelDownload = false;

	if( szTempDir )
	{
		m_szTempDir = (char*) malloc(strlen(szTempDir) + 2 );
		memset( m_szTempDir, 0, strlen(szTempDir) + 2 );
		memcpy( m_szTempDir, szTempDir, strlen(szTempDir) );
		if( m_szTempDir[strlen(m_szTempDir)-1] != '/')
		{
			m_szTempDir[strlen(m_szTempDir)] = '/';
		}
	}
	else
	{
		m_szTempDir = strdup("./");
	}
}

ArticleDownloader::~ArticleDownloader()
{
	if( m_szTempDir )
	{
		free( m_szTempDir );
	}
}

void ArticleDownloader::Run()
{
	// Pre-init
	bool bThrottleDownloadRate = false;
	if( GlobalControl::GetInstance()->GetDownloadRate() > 0.0f )
	{
		bThrottleDownloadRate = true;
	}

//	m_pJob->SetStatus( JOBSTATUS_RUNNING );
	
	int level = 0;
	int iNumberOfServers = MultiServerPool::GetInstance()->GetNumberServers();
	while( !m_bCancelDownload && (level < iNumberOfServers) )
	{
		m_pJob->SetStatus( JOBSTATUS_RUNNING );
		
		NNTPConnection* pConnection = MultiServerPool::GetInstance()->GetConnection( level );

		if(!pConnection)
		{
			// all pConnectionection-levels passed, still not finished...
			// seems like we failed.
			m_pJob->SetStatus(JOBSTATUS_FAILED);
			error("%s FAILED! Article unavailable on all servers.", m_pJob->GetMessageID());
			break;
        	}

		// Okay, we got a pConnectionection. Now start downloading!!

		// but at first, change group! dryan's server wants it this way... ;-)
		bool grpchanged = true;

		for (unsigned int i = 0; i < m_pJob->GetGroups()->size(); i++)
		{
			if (pConnection->JoinGroup((*(m_pJob->GetGroups()))[i]))
			{
				grpchanged = false;
			}
			else
			{
				grpchanged = true;
				break;
			}
		}

		if (grpchanged)
		{
			// now, let's begin!
			char tmp[1024];
      
			snprintf(tmp, 1024, "ARTICLE %s\r\n", m_pJob->GetMessageID());

			char* answer = NULL;

			for (int retry = 3; retry > 0; retry--)
			{
				answer = pConnection->Request(tmp);
				if (answer && (!strncmp(answer, "2", 1)))
				{
					break;
				}
			}

			if (answer && (!strncmp(answer, "2", 1)))
			{
				// positive answer!
				g_pNZBQueueCoordinator->JobStarted( m_pJob, pConnection );
              
				char* line = NULL;

				char tmpfilename[1024];

				snprintf(tmpfilename, 1024, "%snzbgettemp-XXXXXX", m_szTempDir);

				int tmpfd = mkstemp(tmpfilename);

				FILE* outfile = fdopen(tmpfd, "w+");
				if (outfile)
				{

					bool body = false;
					int lines = 0;
					int maxlines = 0;
					int bytes = 0;

					while( !m_bCancelDownload )
					{
						SetLastUpdateTimeNow();
					
						// Throttle the bandwidth
						while( bThrottleDownloadRate && g_pNZBQueueCoordinator->CalcCurrentDownloadSpeed() > GlobalControl::GetInstance()->GetDownloadRate() )
						{
							SetLastUpdateTimeNow();
							usleep( 500 * 1000 );
						}

						// Global Pausing??
						while( GlobalControl::GetInstance()->GetPause() )
						{
							SetLastUpdateTimeNow();
							usleep( 1000 * 1000 );
						}

						line = pConnection->ReadLine();

						// Have we encountered a timeout?
						if( (int)line == EOF )
						{
							perror( "pConnection->ReadLine() encountered error" );
							m_pJob->SetStatus(JOBSTATUS_FAILED);
							break;
							
						}

						//detect end of article
						if ((!line) || (!strcmp(line, ".\r\n")) || (!strcmp(line, ".\n")))
						{
							break;
						}

						// Did we meet an unexpected need for authorization?
						if( !strncmp(line, "480", 3) )
						{
							pConnection->Authenticate();
						}

						//detect lines starting with "." (marked as "..")
						if (!strncmp(line, "..", 2))
							line++;
                      
						// detect Lines-header
						if ((!maxlines) && (!strncmp(line, "Lines: ",7)))
							maxlines = atoi(line+7);
                      
						// count lines in body
						if (body)
							lines++;
                      
						// detect end og header
						if ((!body) && ((!strcmp(line,"\r\n")) || (!strcmp(line,"\n"))))
						{
							body = true;
						}
                      
						int wrcnt = fwrite(line, 1, strlen(line), outfile);
						if (wrcnt > 0)
						{
							bytes += wrcnt;
						}
                                            
						if ((lines % 100 == 0) || (lines == maxlines))
						g_pNZBQueueCoordinator->JobProgress( m_pJob, pConnection, bytes, lines, maxlines );
					}
                  
					fflush(outfile);
					fclose(outfile);
                  
					if (line)
					{
						g_pNZBQueueCoordinator->JobFinished( m_pJob, pConnection );
						MultiServerPool::GetInstance()->FreeConnection(level ,pConnection);

						struct timeval StartTime, EndTime;
						gettimeofday( &StartTime, 0 );
						char* resFilename = Decoder::GetInstance()->Decode(tmpfilename, m_szTempDir);
						gettimeofday( &EndTime, 0 );

						float fDeltaTime = ( (EndTime.tv_sec-StartTime.tv_sec) + (EndTime.tv_usec-StartTime.tv_usec) / 1000000 );

//						info( "UUDecode in %f seconds", fDeltaTime );

						if (resFilename)
						{
							info("Decoded %s - Decode time %.0f second(s)", resFilename, fDeltaTime);
							
							m_pJob->SetResultFilename(resFilename);
							m_pJob->SetStatus(JOBSTATUS_FINISHED);
							
							free(resFilename);
							remove(tmpfilename);
							break;
						}
						else
						{
							error("Decoding %s failed", m_pJob->GetMessageID());
							// possibly permanent error, abort
							m_pJob->SetStatus(JOBSTATUS_FAILED);
							break;
						}
                    }
                  else
                    {
                      warn("Unexpected end of %s", m_pJob->GetMessageID());
					  m_pJob->SetStatus( JOBSTATUS_FAILED );  // Set as failed
					}

                  remove(tmpfilename);

                }
              else
                {
                  warn("Could not create file %s", tmpfilename);
                }
          
            }
          else
            {
              if (!answer)
                warn("%s @ %s failed: Connection closed by remote host.", m_pJob->GetMessageID(), pConnection->GetServer()->GetHost());
              else
                warn("%s @ %s failed (%s)", m_pJob->GetMessageID(), pConnection->GetServer()->GetHost(), answer);
              g_pNZBQueueCoordinator->JobFinished( m_pJob, pConnection );

			  m_pJob->SetStatus( JOBSTATUS_FAILED );  // Set as failed
            }
        }
      else
        {
          //group-change failed. Ignore this, too noisy to report
			m_pJob->SetStatus( JOBSTATUS_FAILED );
        }
      
		MultiServerPool::GetInstance()->FreeConnection(level,pConnection);
      g_pNZBQueueCoordinator->JobFinished( m_pJob, pConnection );
      
      level++;
    }

  Stop();
}
