/*
 *  This file if part of nzbget
 *
 *  Copyright (C) 2004  Sven Henkel <sidddy@users.sourceforge.net>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Revision: 1.13 $
 * $Date: 2005/07/12 09:37:48 $
 *
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string.h>
#include <stdlib.h>
#include <netdb.h>
#include <unistd.h>

#include "global.h"
#include "Connection.h"
#include "Options.h"

#define MIN(a,b) ((a>b)?b:a)
#define MAX(a,b) ((a>b)?a:b)

extern Options* g_pOptions;

Connection::Connection( NetAddress* pNetAddress )
{
	debug("Connection::Connection");
	m_pNetAddress		= pNetAddress;
	m_iStatus			= CONNECTION_STATUS_DISCONNECTED;
	m_iBufferLength		= 0;
	m_iBufferPos		= 0;
	m_pFileBuffer		= NULL;
	m_iSocket			= -1;

	m_pReadBuffer		= (char*)malloc( CONNECTION_BUFFER_SIZE+1 );
	m_pReadBuffer[CONNECTION_BUFFER_SIZE] = 0;
}

Connection::~Connection()
{
	debug("Connection::~Connection()");
	if( m_iStatus == CONNECTION_STATUS_CONNECTED )
	{
		Disconnect();
	}

	if( m_pReadBuffer )
	{
		free( m_pReadBuffer );
	}
}

int Connection::Connect()
{
	debug("Connection::connect()");
	if( m_iStatus == CONNECTION_STATUS_CONNECTED )
		return 0;
	
	int iRes = DoConnect();
	
	if( iRes >= 0 )
		m_iStatus = CONNECTION_STATUS_CONNECTED;
	else
		Connection::DoDisconnect();
	
	return iRes;
}

int Connection::Disconnect()
{
	debug("Connection::disconnect()");
	if( m_iStatus == CONNECTION_STATUS_DISCONNECTED )
		return 0;
	
	int iRes = DoDisconnect();
	
	m_iStatus = CONNECTION_STATUS_DISCONNECTED;
	
	return iRes;
}

int Connection::Bind()
{
	debug("Connection::disconnect()");

	if( m_iStatus == CONNECTION_STATUS_LISTENING )
	{
		return 0;
	}

	int iRes = DoBind();

	m_iStatus = CONNECTION_STATUS_LISTENING;

	return iRes;
}

int Connection::Write(char* line)
{
	debug("Connection::write(char* line)");

	if( m_iStatus == CONNECTION_STATUS_DISCONNECTED )
	{
		int iRes = Connect();
		if( iRes < 0 )
		return iRes;
	}
	
	int iRes = DoWrite(line);
	
	if( iRes == EOF )
		Connection::DoDisconnect();
	
	return iRes;
}

int Connection::SendBuffer( void* pData, int iLength )
{
	return SendBuffer( m_iSocket, pData, iLength );
}

int Connection::SendBuffer( int iSocket, void* pData, int iLength )
{
	if( m_iStatus == CONNECTION_STATUS_DISCONNECTED )
	{
		int iRes = Connect();
		if( iRes < 0 )
		{
			return iRes;
		}
	}

	int l;
	char* p = (char*)pData;
	int iRes;

	do
	{
		l = MIN( iLength, CONNECTION_BUFFER_SIZE );

		iRes = Send( iSocket, (char*)p, l );

		iLength -= CONNECTION_BUFFER_SIZE;

		p += CONNECTION_BUFFER_SIZE;
	}
	while( iLength >= 0 );

	return iRes;
}

int Connection::Send( char* pData, int iLength )
{
	debug( "Connection::Send" );

	if( m_iStatus == CONNECTION_STATUS_DISCONNECTED )
	{
		int iRes = Connect();
		if( iRes < 0 )
		return iRes;
	}

	int iRes = Send( m_iSocket, pData, iLength );
	
	return iRes;
}

int Connection::Send( int iSocket, char* pData, int iLength )
{
	debug( "Connection::Send" );

	int iRes = send( iSocket, pData, iLength, 0 );
	
	return iRes;
}



char* Connection::ReadLine()
{
	if( m_iStatus == CONNECTION_STATUS_DISCONNECTED )
	{
		int iRes = Connect();
		if( iRes < 0 )
			return NULL;
	}
	
	char* res = DoReadLine();
	
	if (res == NULL)
		Connection::DoDisconnect();
	
	return res;
}

int Connection::Accept()
{
	if( m_iStatus != CONNECTION_STATUS_LISTENING )
	{
		int iRes = Bind();
		if( iRes < 0 )
		{
			return NULL;
		}
	}

	int iRes = DoAccept();

	return iRes;
}

int Connection::Recv( int iSocket )
{
	return Recv( iSocket, m_pReadBuffer, CONNECTION_BUFFER_SIZE );
}

int Connection::Recv( int iSocket, char* pBuffer, int iSize )
{
	memset( pBuffer, 0, iSize );
	
	int iReceived = recv( iSocket, pBuffer, iSize, 0 );

	if( iReceived < 0 )
	{
		perror( "recv" );
	}

	return iReceived;
}

int Connection::DoConnect()
{
	debug("Connection::doConnect()");
	
	//resolve hostname
	m_iBufferLength = 0;
	m_iBufferPos = 0;

	struct hostent *hinfo;
	
	hinfo = ::gethostbyname( m_pNetAddress->GetHost() );
	
	if( !hinfo )
	{
		error( "Could not resolve hostname %s", m_pNetAddress->GetHost() );
		return -1;
	}
	
	memset(&m_SocketAddress, '\0', sizeof(m_SocketAddress));
	m_SocketAddress.sin_family = AF_INET;
	memcpy(&m_SocketAddress.sin_addr.s_addr, hinfo->h_addr_list[0], sizeof(m_SocketAddress.sin_addr.s_addr));
	m_SocketAddress.sin_port = htons( m_pNetAddress->GetPort() );
	
	m_iSocket = ::socket(PF_INET,SOCK_STREAM,0);
	
	if( m_iSocket < 0 )
	{
		error("Socket creation failed for %s!", m_pNetAddress->GetHost() );
		return -1;
	}

	int res = ::connect( m_iSocket ,(struct sockaddr *) &m_SocketAddress, sizeof(m_SocketAddress) );
	
	if (res < 0)
	{
		error("Connection to %s failed!", m_pNetAddress->GetHost() );
		return -1;
	}
	
	m_pFileBuffer = fdopen( m_iSocket, "r+" );
	
	if( !m_pFileBuffer )
	{
		error( "fdopen failed for %s", m_pNetAddress->GetHost() );
		return -1;
	}

	struct timeval TimeVal;

	m_fDownloadTimeout      = g_pOptions ? atof( g_pOptions->getOption( OPTION_DOWNLOADTIMEOUT, NULL ) ) : 60.0f;

	TimeVal.tv_sec          = (long int) m_fDownloadTimeout;
	TimeVal.tv_usec         = 0;

	setsockopt( m_iSocket, SOL_SOCKET, SO_RCVTIMEO, &TimeVal, sizeof(TimeVal) );
	
	return 0;
}

int Connection::DoDisconnect()
{
	debug("Connection::doDisconnect()");

	if( m_pFileBuffer )
		fclose( m_pFileBuffer );

	if( m_iSocket > -1 )
		close( m_iSocket );
	
	m_iSocket		= -1;
	m_pFileBuffer	= NULL;
	m_iStatus		= CONNECTION_STATUS_DISCONNECTED;
	
	return 0;
}

int Connection::DoWrite( char* szText )
{
	debug("Connection::doWrite(char* text)");
	return ::fputs( szText, m_pFileBuffer );
}

char* Connection::DoReadLine()
{
	debug( "Connection::DoReadLine()" );
	return ::fgets( m_pReadBuffer, CONNECTION_BUFFER_SIZE, m_pFileBuffer );
}

int Connection::DoBind()
{
	debug("Connection::DoBind()");
	
	struct hostent *hinfo;
	
	hinfo = ::gethostbyname( m_pNetAddress->GetHost() );
	
	if( !hinfo )
	{
		error( "Could not resolve hostname %s", m_pNetAddress->GetHost() );
		return -1;
	}
	
	m_iSocket = ::socket( PF_INET, SOCK_STREAM, 0 );
	
	if( m_iSocket < 0 )
	{
		error("Socket creation failed for %s!", m_pNetAddress->GetHost() );
		return -1;
	}
	
	memset(&m_SocketAddress, '\0', sizeof(m_SocketAddress));
	m_SocketAddress.sin_family = AF_INET;
	m_SocketAddress.sin_addr.s_addr = htonl(INADDR_ANY);
	m_SocketAddress.sin_port = htons( m_pNetAddress->GetPort() );

	if( bind( m_iSocket, (struct sockaddr *) &m_SocketAddress, sizeof(m_SocketAddress)) < 0)
	{
		error( "binding socket failed for %s!", m_pNetAddress->GetHost() );
		return -1;
	}

	if( listen( m_iSocket, 10 ) < 0 )
	{
		error( "listen on socket failed for %s!", m_pNetAddress->GetHost() );
		return -1;
	}

/*
	m_pFileBuffer = fdopen( m_iSocket, "r+" );
	
	if( !m_pFileBuffer )
	{
		error( "fdopen failed for %s", m_pNetAddress->GetHost() );
		return -1;
	}
*/

	return 0;
}

int Connection::DoAccept()
{
	struct sockaddr_in ClientAddress;
	socklen_t SockLen;

	SockLen = sizeof( ClientAddress );

	int iSocket = accept( GetSocket(), (struct sockaddr *)&ClientAddress, &SockLen );

	if( iSocket < 0 )
	{
		error( "cannot accept connection" );
		return -1;
	}

	return iSocket;
}

void Connection::FlushFileBuffer()
{
	fflush( m_pFileBuffer );
}

const char* Connection::GetStatusString()
{
	switch( m_iStatus )
	{
		case CONNECTION_STATUS_CONNECTED:
			return "Connected";
		case CONNECTION_STATUS_DISCONNECTED:
			return "Disconnected";
		case CONNECTION_STATUS_LISTENING:
			return "Listening";
	}

	return "N/A";
}
