/*
 *  This file if part of nzbget
 *
 *  Copyright (C) 2004  Sven Henkel <sidddy@users.sourceforge.net>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Revision: 1.30 $
 * $Date: 2005/07/12 09:37:48 $
 *
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <vector>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <dirent.h>
#include <fcntl.h>

#include "FileDownloader.h"
#include "ArticleDownloader.h"
#include "ArticleJob.h"
#include "global.h"
#include "GlobalControl.h"
#include "Options.h"
#include "Observable.h"
#include "MultiServerPool.h"

extern Options* g_pOptions;

FileDownloader::FileDownloader(DownloadJob* pJob, char* tempdir, char* destdir)
{
	m_bCancelDownload = false;

	m_fDownloadTimeout = 360.0f;
	
	m_pJob = pJob;
	m_ArticleJobs.clear();
	m_ArticleDownloads.clear();
	
	if( tempdir )
	{
		m_szTempDir = (char*) malloc(strlen(tempdir) + 2);
		memset(m_szTempDir, 0, strlen(tempdir) + 2);
		memcpy(m_szTempDir, tempdir, strlen(tempdir));
		if (m_szTempDir[strlen(m_szTempDir)-1] != '/')
		{
			m_szTempDir[strlen(m_szTempDir)] = '/';
		}
	}
  	else
	{
		m_szTempDir = strdup("./");
	}

	if (destdir)
	{
		m_szDestDir = (char*) malloc(strlen(destdir) + 2);
		memset(m_szDestDir, 0, strlen(destdir) + 2);
		memcpy(m_szDestDir, destdir, strlen(destdir));
		
		if (m_szDestDir[strlen(m_szDestDir)-1] != '/')
		{
			m_szDestDir[strlen(m_szDestDir)] = '/';
		}
	}
	else
	{
		m_szDestDir = strdup("./");
	}

	// Ensure the destination dir is created
	mkdir(m_szDestDir, S_IRUSR | S_IWUSR | S_IXUSR | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH);

	// Ensure the tmp dir is created
	mkdir(m_szTempDir, S_IRUSR | S_IWUSR | S_IXUSR | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH);

	pthread_mutex_init( &m_ArticleDownloadMutex, NULL );
}

FileDownloader::~FileDownloader()
{
	if( m_szTempDir )
		free(m_szTempDir);
	
	if( m_szDestDir )
		free(m_szDestDir);
	
	for( unsigned int i = 0; i < m_ArticleJobs.size(); i++ )
	{
		if( m_ArticleJobs[i] )
		{
			delete m_ArticleJobs[i];
			m_ArticleJobs[i] = NULL;
		}
	}
	
	for( unsigned int i = 0; i < m_ArticleDownloads.size(); i++ )
	{
		if( m_ArticleDownloads[i] )
		{
			delete m_ArticleDownloads[i];
			m_ArticleJobs[i] = NULL;
		}
	}
}


void FileDownloader::Start()
{

	if( IsDupe() )
	{
		m_pJob->SetStatus(JOBSTATUS_FINISHED);
		warn("File with subject \"%s\" seems to be duplicate, skipping.",m_pJob->getSubject());
		return;
	}

	for (unsigned int i = 0; i < m_pJob->getMessageIDs()->size(); i++)
	{
		if ((*(m_pJob->getMessageIDs()))[i] != NULL)
		{
			//start only when level-0-connection is available
			NNTPConnection* conn = MultiServerPool::GetInstance()->GetConnection(0);
			MultiServerPool::GetInstance()->FreeConnection(0, conn);
	
			ArticleJob* cjob = new ArticleJob((*(m_pJob->getMessageIDs()))[i]);
	
			for (unsigned int j = 0; j < m_pJob->getGroups()->size(); j++)
			{
				cjob->GetGroups()->push_back((*(m_pJob->getGroups()))[j]);
			}
	
			cjob->AddObserver(this);
			m_ArticleJobs.push_back(cjob);
		
			ArticleDownloader* dl = new ArticleDownloader(cjob, m_szTempDir);
	
			m_ArticleDownloads.push_back(dl);
		}
	}
	
	for (unsigned int i = 0; i < m_ArticleDownloads.size(); i++)
	{
		//start only when free connections are available for every level
		int level = 0;

		while( true )
		{
			NNTPConnection* conn = MultiServerPool::GetInstance()->GetConnection(level);
			if (conn)
			{
				MultiServerPool::GetInstance()->FreeConnection(level, conn);
			}
			else
			{
				break;
			}
			level++;
		}

		if( m_bCancelDownload )
		{
			return;
		}
	
		m_ArticleDownloads[i]->Start();
		usleep(100*1000);
	}

	// TEST
/*
	int iNotFinished = 0;
	for (unsigned int i = 0; i < m_ArticleJobs.size(); i++)
	{
		int iStatus = m_ArticleJobs[i]->getStatus();
		if((iStatus != JOBSTATUS_FINISHED)&& (iStatus != JOBSTATUS_FAILED))
		{
			iNotFinished++;
		}
	}

	if( iNotFinished > 0 )
	{
		info( "Download finished but %i jobs hanging", iNotFinished );
	}
*/
}

void FileDownloader::Notify(Observable* caller)
{
	debug("FileDownloader::notify(Observable* caller)");
	
	// all child-jobs ready?
	for (unsigned int i = 0; i < m_ArticleJobs.size(); i++)
	{
		if((m_ArticleJobs[i]->getStatus() != JOBSTATUS_FINISHED)&& (m_ArticleJobs[i]->getStatus() != JOBSTATUS_FAILED))
		{
			return;
		}
	}
	
	// all ready!
	
	// determine filename
	
	char* basename = NULL;
	int maxpart = 0;
	
	for (unsigned int i = 0; i < m_ArticleJobs.size(); i++)
	{
		char* partfn = m_ArticleJobs[i]->GetResultFilename();
		if (partfn)
		{
			char* lastdot = strrchr(partfn, '.');
	
			if (lastdot)
			{
				char* localbasename = (char*) malloc(lastdot-partfn+1);
				memset(localbasename, 0, lastdot-partfn+1);
				memcpy(localbasename, partfn, lastdot-partfn);
	
				if (strncmp(localbasename, "nzbgettemp-unknown-", 19))
				{
					if (basename)
					{
						if (strcmp(basename, localbasename))
						{
							error("Filenames differ in parts of post!");
							m_pJob->SetStatus(JOBSTATUS_FAILED);
							return;
						}
						else
						{
							free(localbasename);
						}
					}
					else
					{
						basename = localbasename;
					}
				}
	
				int localpart = atoi(lastdot+1);
				if (localpart > maxpart)
				{
					maxpart = localpart;
				}
			}
		}
	}
	
	if (!basename)
	{
		//ooh, something f*cked up, so let's create our own name
		basename = (char*) malloc(1024);
		snprintf(basename, 1024, "nzbget-unknown-XXXXXX");
		mktemp(basename);
	}
	
	// search childs for unnamed parts
	for (unsigned int i = 0; i < m_ArticleJobs.size(); i++)
	{
		char* partfn = m_ArticleJobs[i]->GetResultFilename();
		if (partfn)
		{
			if (!strncmp(partfn, "nzbgettemp-unknown-", 19))
			{
				char oldfn[1024];
				char newfn[1024];
	
				snprintf(oldfn, 1024, "%s%s", m_szTempDir, partfn);
				snprintf(newfn, 1024, "%s%s.%d", m_szTempDir, basename, i+1);
	
				rename(oldfn, newfn);
	
				if ((signed int)i+1 > maxpart)
					maxpart = i+1;
			}
		}
	}
	
	FILE* outfile;
	
	char ofn[1024];
	snprintf(ofn, 1024, "%s%s",m_szDestDir, basename);

	// prevent overwriting existing files
	struct stat statbuf;
	int dupcount = 0;
	while (!stat(ofn, &statbuf))
	{
		dupcount++;
		snprintf(ofn, 1024, "%s%s_duplicate%d",m_szDestDir, basename, dupcount);
	}
	
	outfile = fopen(ofn, "w+");
	
	if (!outfile)
	{
		error("Could not create file %s!", basename);
		m_pJob->SetStatus(JOBSTATUS_FAILED);
		return;
	}
	
	
	bool complete = true;
	int iBrokenCount = 0;
	
	for (int i = 1; i <= maxpart; i++)
	{
		FILE* infile;
	
		char fn[1024];
		snprintf(fn, 1024, "%s%s.%d", m_szTempDir, basename, i);
	
		infile = fopen(fn, "r");
	
		if (infile)
		{
			char buffer[10240];
	
			int cnt = 10240;
	
			while (cnt == 10240)
			{
				cnt = fread(buffer, 1, 10240, infile);
				fwrite(buffer, 1, cnt, outfile);
			}
	
			fclose(infile);
			remove(fn);
		}
		else
		{
			complete = false;
			iBrokenCount++;
			nzblog( "Could not find file %s. Status is broken", fn );
		}
	}
	
	fclose(outfile);
	
	if (complete)
	{
		nzblog( "Successfully downloaded %s", ofn );
		m_pJob->SetStatus(JOBSTATUS_FINISHED);
	//      info("Decoded %s", ofn);
	}
	else
	{
		if (strcmp(g_pOptions->getOption(OPTION_RENAMEBROKEN, "yes"), "yes") == 0)
		{
			char brokenfn[1024];
			snprintf(brokenfn, 1024, "%s_broken", ofn);
			rename(ofn, brokenfn);

			nzblog( "Renaming broken file from %f to %f", ofn, brokenfn );
		}
		else
		{
			nzblog( "Not renaming broken file %s", ofn );
		}

		if( GlobalControl::GetInstance()->GetCreateBrokenFileLog() )
		{
			char szBrokenLogName[1024];
			snprintf( szBrokenLogName, 1024, "%s/_brokenlog.txt",m_szDestDir);

			LogBrokenFile( szBrokenLogName, basename, maxpart, iBrokenCount );
		}

		m_pJob->SetStatus(JOBSTATUS_FAILED);
		warn("%s is incomplete!", ofn);
	}
}


bool FileDownloader::IsRunning()
{
	for (unsigned int i = 0; i < m_ArticleDownloads.size(); i++)
	{
		if (m_ArticleDownloads[i]->IsRunning())
		{
			return true;
		}
	}
	return false;
}


bool FileDownloader::IsDupe()
{

	if (strcmp(g_pOptions->getOption(OPTION_DUPECHECK, "yes"), "yes"))
	{
		return false;
	}
	
	if ((!m_pJob) || (!m_pJob->getSubject()) || (!m_szDestDir))
	{
		return false;
	}
	
	
	DIR *dir=opendir(m_szDestDir);
	struct dirent *entry;
	if (!dir)
	{
		return false;
	}
	
	std::vector< char* > existing;
	
	existing.clear();
		
	while ((entry=readdir(dir)))
	{
		if (!strcmp(entry->d_name, ".."))
			continue;
	
		if (!strcmp(entry->d_name,"."))
			continue;
	
		if (entry->d_name[0] == '.')
			continue;
	
		// do not consider short filenames, too many false positives!!
		if ((strlen(entry->d_name) < 6) && (strstr(entry->d_name, ".") == NULL))
			continue;
	
		existing.push_back(strdup(entry->d_name));
	}
	
	closedir(dir);
	
	// mmkay, got the names of existing files, now make substring-tests
	// with subject
	
	bool res = false;
	
	for (unsigned int i = 0; i < existing.size(); i++)
	{
		if (strstr(m_pJob->getSubject(), existing[i]))
		{
			// found a match!
			res = true;
		}
		free(existing[i]);
		existing[i] = NULL;
	}
	
	existing.clear();
	return res;
}

void FileDownloader::CancelDownload()
{
	m_bCancelDownload = true;

	std::vector<ArticleDownloader*>::iterator it;
	for( it=m_ArticleDownloads.begin(); it!=m_ArticleDownloads.end(); it++ )
	{
		(*it)->CancelDownload();
	}
}

void FileDownloader::CleanupArticleDownloads()
{
	pthread_mutex_lock( &m_ArticleDownloadMutex );
	debug( "FileDownloader::CleanupArticleDownloads() (lock)" );
	
	std::vector<ArticleDownloader*>::iterator it = m_ArticleDownloads.begin();
	while( it != m_ArticleDownloads.end() )
	{
		ArticleDownloader* pArticleDownloader = *it;
		switch( pArticleDownloader->GetArticleJob()->getStatus() )
		{
			case JOBSTATUS_ASSIGNED:
			case JOBSTATUS_RUNNING:
			{
				// how long since the thread was last active?
				time_t tDeltaTime = ::time( NULL ) - pArticleDownloader->GetLastUpdateTime();
				if( tDeltaTime > m_fDownloadTimeout )
				{
					// Printout info
					char* szName = pArticleDownloader->GetArticleJob()->GetResultFilename();
					warn( "Aborting download of %s due to unactivity (inactive for more than %f seconds)", szName, m_fDownloadTimeout );
					
					// Free the connection
//					MultiServerPool::GetInstance()->FreeConnection( pArticleDownloader->GetLevel(), pArticleDownloader->GetConnection() );

					// Stop the thread
#ifndef DEBUG_SINGLE_THREADED_DOWNLOAD
					pthread_cancel( pArticleDownloader->GetThread() );
#endif
					
					// Remove the articledownloader from vector
					m_ArticleDownloads.erase( it );
					continue;
				}
				break;
				
			}
//			case JOBSTATUS_DECODING:
//				// Do nothing if waiting for decoding
//				break;
			case JOBSTATUS_FINISHED:
			case JOBSTATUS_FAILED:
			{
				// Stop the thread
#ifndef DEBUG_SINGLE_THREADED_DOWNLOAD
				pthread_cancel( pArticleDownloader->GetThread() );
#endif
				// Remove the articledownloader
				m_ArticleDownloads.erase( it );
				continue;
			}
			default:
				break;
		}

		it++;
	}
    
	debug( "FileDownloader::CleanupArticleDownloads() (unlock)" );
	pthread_mutex_unlock( &m_ArticleDownloadMutex );
}

void FileDownloader::PlotArticleDownloads()
{
	pthread_mutex_lock( &m_ArticleDownloadMutex );
	debug( "FileDownloader::PlotArticleDownloads() (lock)" );
	
	std::vector<ArticleDownloader*>::iterator it;
	for( it=m_ArticleDownloads.begin(); it!=m_ArticleDownloads.end(); it++ )
	{
/*		ArticleDownloader* pArticleDownloader = *it;

		ArticleJob* pArticleJob = pArticleDownloader->GetArticleJob();
//		float fTime = ::time( NULL );
		switch( pArticleJob->GetStatus() )
		{
			case JOBSTATUS_ASSIGNED:
			{
				time_t tDeltaTime = ::time( NULL ) - pArticleDownloader->GetLastUpdateTime();
				debug( "ArticleJob(Assigned)(%p) %s: Delta update time %i", pArticleDownloader, pArticleJob->getMessageID(), tDeltaTime );
				break;
			}
			case JOBSTATUS_RUNNING:
			{
				time_t tDeltaTime = ::time( NULL ) - pArticleDownloader->GetLastUpdateTime();
				debug( "ArticleJob(Running)(%p) %s: Delta update time %i", pArticleDownloader, pArticleJob->getMessageID(),tDeltaTime );
				break;
			}
			case JOBSTATUS_FINISHED:
			{
				debug( "ArticleJob(Finished)(%p) %s", pArticleDownloader, pArticleJob->getMessageID() );
				break;
			}
			case JOBSTATUS_FAILED:
			{
				debug( "ArticleJob(Failed)(%p) %s", pArticleDownloader, pArticleJob->getMessageID() );
				break;
			}
			default:
				debug( "ArticleJob(UNKNOWN)(%p) %s", pArticleDownloader, pArticleJob->getMessageID() );
				break;
		}
		*/
	}

	debug( "FileDownloader::PlotArticleDownloads() (unlock)" );
	pthread_mutex_unlock( &m_ArticleDownloadMutex );
}

void FileDownloader::LogBrokenFile( const char* szBrokenLogName, const char* szFileName, int iParts, int iBroken )
{
	FILE* file = fopen( szBrokenLogName, "a+" );
	float fBrokenPrc = 1.0f - (float)(iBroken) / (float)iParts;
	fprintf( file, "%s (%.0f\%)\n", szFileName, 100.0f * fBrokenPrc );
	fclose( file );
}
