/*
 *  This file if part of nzbget
 *
 *  Copyright (C) 2004  Sven Henkel <sidddy@users.sourceforge.net>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Revision: 1.17 $
 * $Date: 2005/07/12 09:37:48 $
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <stdarg.h>

#include "Frontend.h"
#include "GlobalControl.h"

#include "NZBQueueCoordinator.h"

extern int g_iThreadCount;
extern NZBQueueCoordinator* g_pNZBQueueCoordinator;

Frontend::Frontend()
{
	m_pDownloadSpeedMonitor = new DownloadSpeedMonitor();
	m_Messages.clear();
	m_lQueueSize = 0;
	
	pthread_mutex_init( &m_MutexMessages, NULL );

	
}

Frontend::~Frontend()
{
}

void Frontend::FrontendError(char* msg)
{
	char tmp[1024];
	
	snprintf(tmp, 1024, "[ERROR] %s\n", msg);
	pthread_mutex_lock(&m_MutexMessages);
	m_Messages.push_back(strdup(tmp));
	pthread_mutex_unlock(&m_MutexMessages);
}

void Frontend::FrontendWarn(char* msg)
{
	char tmp[1024];
	
	snprintf(tmp, 1024, "[WARNING] %s\n", msg);
	pthread_mutex_lock(&m_MutexMessages);
	m_Messages.push_back(strdup(tmp));
	pthread_mutex_unlock(&m_MutexMessages);
}

void Frontend::FrontendDebug(char* msg)
{
	char tmp[1024];
	
	snprintf(tmp, 1024, "[DEBUG] %s\n", msg);
	pthread_mutex_lock(&m_MutexMessages);
	m_Messages.push_back(strdup(tmp));
	pthread_mutex_unlock(&m_MutexMessages);
}

void Frontend::FrontendInfo(char* msg)
{
	char tmp[1024];
	
	snprintf(tmp, 1024, "[INFO] %s\n", msg);
	pthread_mutex_lock(&m_MutexMessages);
	m_Messages.push_back(strdup(tmp));
	pthread_mutex_unlock(&m_MutexMessages);
}

void Frontend::Start()
{
	Thread::Start();

	if( m_pDownloadSpeedMonitor )
	{
		m_pDownloadSpeedMonitor->Start();
	}
}

void Frontend::Stop()
{
	Thread::Stop();

	if( m_pDownloadSpeedMonitor )
	{
		m_pDownloadSpeedMonitor->Stop();
	}
}

void Frontend::Run()
{
	while( !GlobalControl::GetInstance()->GetQuit() )
	{
		PrintMessages();
		usleep( (int)GetSleepTime() );
	}
}

void Frontend::PrintMessages()
{
	pthread_mutex_lock(&m_MutexMessages);
	
	int newline = 0;
	
	for (unsigned int i = 0; i < m_Messages.size(); i++)
	{
		fprintf(stdout, "%s", m_Messages[i]);
		free(m_Messages[i]);
		m_Messages[i] = NULL;
		newline = 1;
	}
	
	m_Messages.clear();
	
	pthread_mutex_unlock(&m_MutexMessages);
	
	if (newline == 1)
		PrintStatus();
}

void Frontend::PrintStatus()
{
}

void Frontend::FrontendQueueListEntry( const char* szEntry )
{
	char tmp[1024];
	
	snprintf(tmp, 1024, "[ENTRY] %s\n", szEntry);
	pthread_mutex_lock(&m_MutexMessages);
	m_Messages.push_back(strdup(tmp));
	pthread_mutex_unlock(&m_MutexMessages);
}


//--------------------------------------------------------------------------------------------
//////////////////////////////////////////////////////////////////////////////////////////////
//--------------------------------------------------------------------------------------------


DownloadSpeedMonitor::DownloadSpeedMonitor()
{
	ClearStats();
}

void DownloadSpeedMonitor::Run()
{
	while( !GlobalControl::GetInstance()->GetQuit() )
	{
		usleep( GetSleeptime() );
		UpdateDownloadSpeed();
	}
}

void DownloadSpeedMonitor::UpdateDownloadSpeed()
{
	float fCurrentDownloadSpeed = g_pNZBQueueCoordinator->CalcCurrentDownloadSpeed();

	// Only push 
	if( fCurrentDownloadSpeed <= 0.0f ) return;

	// Push on stack
	m_Seconds10.push_back( fCurrentDownloadSpeed );

	// Have we filled the stack?
	if( !m_Seconds10.size() < 6 ) return;

	float fAverage = 0.0f;
	std::vector<float>::iterator it;
	for( it=m_Seconds10.begin(); it != m_Seconds10.end(); it++ )
	{
		fAverage += *it;
	}

	fAverage /= m_Seconds10.size();
	m_Seconds10.clear();

	m_Minutes.push_back( fAverage );
}

void DownloadSpeedMonitor::ClearStats()
{
	m_Seconds10.clear();
	m_Minutes.clear();
	m_Minutes10.clear();
	m_Hours.clear();
}
