/*
 *  This file if part of nzbget
 *
 *  Copyright (C) 2004  Sven Henkel <sidddy@users.sourceforge.net>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Revision: 1.11 $
 * $Date: 2005/07/12 09:37:48 $
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <stdarg.h>
#include <signal.h>

#include "NCursesFrontend.h"

#include "NZBQueueCoordinator.h"

#include "GlobalControl.h"

extern int g_iThreadCount;
extern NZBQueueCoordinator* g_pNZBQueueCoordinator;

static const float fNZBQueueScreenPercentage	= 0.6f;
static const unsigned int iMaxNumberMessages	= 40;

#define NCURSES_COLORPAIR_BORDER	1
#define NCURSES_COLORPAIR_TEXT		2
#define NCURSES_COLORPAIR_INFO		3
#define NCURSES_COLORPAIR_WARNING	4
#define NCURSES_COLORPAIR_ERROR		5
#define NCURSES_COLORPAIR_DEBUG		6
#define NCURSES_COLORPAIR_STATUS	7
#define NCURSES_COLORPAIR_KEYBAR	8
#define NCURSES_COLORPAIR_INFOLINE	9
#define NCURSES_COLORPAIR_TEXTHIGHL	10
#define NCURSES_COLORPAIR_CURSOR	11

#define NCOLORON(color) if( m_bUseColor ) attron(COLOR_PAIR(color))
#define NCOLOROFF(color) if( m_bUseColor ) attroff(COLOR_PAIR(color))

#define PADCHAR ' '

NCursesFrontend::NCursesFrontend() : 
m_iLineNr( 0 ),
m_iNrRows( 0 ),
m_iNrColumns( 0 ),
m_iSelectedQueueEntry( 0 ),
m_iInputNumberIndex( 0 ),
m_pWindow( NULL ),
m_eInputMode( eNormal )
{
	// Setup curses
	m_pWindow = initscr();

	if( m_pWindow == NULL )
	{
#ifdef DEBUG
		printf( "ERROR: m_pWindow == NULL\n" );
#endif
		exit( -1 );
	}
	
	m_bUseColor = has_colors();
	if( m_bUseColor )
	{
		start_color();
		init_pair( 0,							COLOR_WHITE, COLOR_BLUE );
		init_pair( NCURSES_COLORPAIR_BORDER,	COLOR_CYAN, COLOR_BLACK );
		init_pair( NCURSES_COLORPAIR_TEXT,		COLOR_WHITE, COLOR_BLACK );
		init_pair( NCURSES_COLORPAIR_INFO,		COLOR_GREEN, COLOR_BLACK );
		init_pair( NCURSES_COLORPAIR_WARNING,	COLOR_MAGENTA, COLOR_BLACK );
		init_pair( NCURSES_COLORPAIR_ERROR,		COLOR_RED, COLOR_BLACK );
		init_pair( NCURSES_COLORPAIR_DEBUG,		COLOR_WHITE, COLOR_BLACK );
		init_pair( NCURSES_COLORPAIR_STATUS,	COLOR_YELLOW, COLOR_BLACK );
		init_pair( NCURSES_COLORPAIR_KEYBAR,	COLOR_WHITE, COLOR_BLUE );
		init_pair( NCURSES_COLORPAIR_INFOLINE,	COLOR_WHITE, COLOR_BLUE );
		init_pair( NCURSES_COLORPAIR_TEXTHIGHL,	COLOR_WHITE, COLOR_RED );
		init_pair( NCURSES_COLORPAIR_CURSOR,	COLOR_BLACK, COLOR_YELLOW );
	}

	m_Messages.clear();
	keypad( stdscr, true );
	nodelay( m_pWindow, true );
	noecho();
	curs_set( 0 ); // Make the cursor invisible
}

NCursesFrontend::~NCursesFrontend()
{
	endwin();
}

void NCursesFrontend::PushOnMessageList( const char* szString, EMessageType eType )
{
	SMessage Msg;
	
	Msg.m_szMessage = strdup(szString);
	Msg.m_eType	= eType;
	
	m_Messages.push_back( Msg );

	// Check if too many entries on list
	if( m_Messages.size() > iMaxNumberMessages )
	{
		std::deque< SMessage >::iterator it = m_Messages.begin();
		
		Msg = *it;
		
		m_Messages.pop_front();
		
		free( Msg.m_szMessage );
	}
}

void NCursesFrontend::FrontendError(char* msg)
{
	char tmp[1024];

	snprintf( tmp, 1024, "%s\n", msg );

	pthread_mutex_lock( &m_MutexMessages );

	PushOnMessageList( tmp, eError );

	pthread_mutex_unlock( &m_MutexMessages );
}

void NCursesFrontend::FrontendWarn(char* msg)
{
	char tmp[1024];

	snprintf(tmp, 1024, "%s\n", msg);

	pthread_mutex_lock(&m_MutexMessages);

	PushOnMessageList( tmp, eWarning );

	pthread_mutex_unlock(&m_MutexMessages);
}

void NCursesFrontend::FrontendDebug(char* msg)
{
	char tmp[1024];

	snprintf(tmp, 1024, "%s\n", msg);

	pthread_mutex_lock(&m_MutexMessages);

	PushOnMessageList( tmp, eDebug );

	pthread_mutex_unlock(&m_MutexMessages);
}

void NCursesFrontend::FrontendInfo(char* msg)
{
	char tmp[1024];

	snprintf(tmp, 1024, "%s\n", msg);

	pthread_mutex_lock(&m_MutexMessages);

	PushOnMessageList( tmp, eInfo );
	
	pthread_mutex_unlock(&m_MutexMessages);
}

void NCursesFrontend::PrintStatus()
{
	char tmp[1024];

	pthread_mutex_lock(&m_MutexMessages);
	
	int iStatusStartRow = (m_iNrRows-1) - 4;

	PlotHeader( "Status", iStatusStartRow, NCURSES_COLORPAIR_INFOLINE );
	iStatusStartRow++;
	ClearRow( iStatusStartRow++ );

	float fCurrentDownloadSpeed = g_pNZBQueueCoordinator->CalcCurrentDownloadSpeed();
	
	char timeString[100];

	timeString[0] = '\0';

	if(fCurrentDownloadSpeed > 0.0)
	{
		long long remain_sec = m_lQueueSize / (fCurrentDownloadSpeed * 1024);
		int h = 0;
		int m = 0;
		int s = 0;
		while (remain_sec > 3600)
		{
			h++;
			remain_sec -= 3600;
		}

		while (remain_sec > 60)
		{
			m++;
			remain_sec -= 60;
		}

		s = remain_sec;

		sprintf(timeString, "~ %.2d:%.2d:%.2d", h, m, s);
	}

	const char* szPause[] = { "Paused", "" };
	int iPauseIdx = GlobalControl::GetInstance()->GetPause() ? 0 : 1;

	char szDownloadLimit[128];
	if( GlobalControl::GetInstance()->GetDownloadRate() > 0.0f )
	{
		sprintf( szDownloadLimit, "Limit %.0f KB/S", GlobalControl::GetInstance()->GetDownloadRate() );
	}
	else
	{
		szDownloadLimit[0] = 0;
	}

	snprintf(tmp, 1024, "%d threads running, %.0f kB/s, %d MB queued (%s) %s %s\n", g_iThreadCount, fCurrentDownloadSpeed, (int) (m_lQueueSize / 1024 / 1024), timeString, szPause[ iPauseIdx ], szDownloadLimit );
	mvprintw( iStatusStartRow++, 0, "%s", tmp );

	ClearRow( iStatusStartRow );

	pthread_mutex_unlock(&m_MutexMessages);
}

void NCursesFrontend::PrintMessages()
{
	pthread_mutex_lock(&m_MutexMessages);

	// Figure out how big the screen is
	int iNrRows, iNrColumns;
	getmaxyx( stdscr, iNrRows, iNrColumns );

	if( iNrRows != m_iNrRows || iNrColumns != m_iNrColumns )
	{
		clear();
		m_iNrRows = iNrRows;
		m_iNrColumns = iNrColumns;

		PrintKeyInputBar();
	}

	m_iLineNr = 0;
	
	//------------------------------------------
	// Print Current NZBQueue
	//------------------------------------------
	PlotHeader( "Queue", m_iLineNr++, NCURSES_COLORPAIR_INFOLINE );
	ClearRow( m_iLineNr++ );

	PrintQueue();

	//------------------------------------------
	// Print Messages
	//------------------------------------------
	ClearRow( m_iLineNr++ );
	PlotHeader( "Messages", m_iLineNr, NCURSES_COLORPAIR_INFOLINE );
	m_iLineNr++;
	ClearRow( m_iLineNr++ );
	
	std::deque< SMessage >::iterator it2;
	
	int iMessageMaxLines = iNrRows - 7;
	unsigned int iLinesLeft = iMessageMaxLines - m_iLineNr;

	uint iCount = 0;	
	
	for( it2 = m_Messages.begin(); it2 != m_Messages.end(); it2++ )
	{
		if( iLinesLeft < m_Messages.size() && iCount++ < (m_Messages.size()-iLinesLeft-1) )
		{
			continue;
		}
	
		SMessage Msg = *it2;
		PrintMessage( Msg, m_iLineNr++ );
		
		if( m_iLineNr > iMessageMaxLines )
		{
			break;
		}
	}

	ClearRow( m_iLineNr++ );
	
	pthread_mutex_unlock(&m_MutexMessages);
	
	PrintStatus();

	// Update the input
	UpdateInput();

	// Cursor placement
	wmove( m_pWindow, m_iNrRows, m_iNrColumns );
	
	// NCurses refresh
	refresh();
}

void NCursesFrontend::OnDownloadCompleted( const char* szFileName )
{
	// Check if this was the last entry which was deleted
	if( g_pNZBQueueCoordinator->GetNZBQueueList().size() == 0 )
	{
		// Force normal mode
		m_eInputMode = eNormal;
		PrintKeyInputBar();
	}
}

void NCursesFrontend::PrintMessage( SMessage& Msg, int iRow )
{
	const char* szMessageType[] = { "[INFO]    ", 
				        "[WARNING] ",
				        "[ERROR]   ", 
				        "[DEBUG]   " };
					
	const int iMessageTypeColor[] = { NCURSES_COLORPAIR_INFO,
					  NCURSES_COLORPAIR_WARNING,
					  NCURSES_COLORPAIR_ERROR,
					  NCURSES_COLORPAIR_DEBUG};

	NCOLORON( iMessageTypeColor[Msg.m_eType] );
	mvprintw( iRow, 0, szMessageType[ Msg.m_eType ] );
	NCOLOROFF( iMessageTypeColor[Msg.m_eType] );

	char szBuffer[512];
	NCOLORON( NCURSES_COLORPAIR_TEXT );
	sprintf( szBuffer, "%s", Msg.m_szMessage );
	int iLength = strlen( szBuffer );
	int iRemain = m_iNrColumns-10-iLength;
	while( iRemain-- >= 0 ) szBuffer[ iLength + iRemain ] = PADCHAR;
	szBuffer[ m_iNrColumns-10-1 ] = 0;
	mvprintw( iRow, 10, "%s", szBuffer );
	NCOLOROFF( NCURSES_COLORPAIR_TEXT );
}

void NCursesFrontend::PlotHeader( const char* szString, int iRow, int iColorPair )
{
	char szBuffer[512];
	
	sprintf( szBuffer, "---| %s |---", szString );
	PlotLine( szBuffer, iRow, iColorPair );
}

void NCursesFrontend::PlotLine( const char* szString, int iRow, int iColorPair )
{
	char szBuffer[512];
	
	sprintf( szBuffer, "%s", szString );
	int iLength = strlen( szBuffer );
	int iRemain = m_iNrColumns - iLength;
	while( iRemain-- > 0 ) szBuffer[ iLength + iRemain ] = PADCHAR;
	szBuffer[ m_iNrColumns-1 ] = 0;
	
	if( m_bUseColor ) attron(COLOR_PAIR(iColorPair));
	mvprintw( iRow, 0, szBuffer );
	if( m_bUseColor ) attroff(COLOR_PAIR(iColorPair));
}

void NCursesFrontend::PrintKeyInputBar()
{
	uint iQueueSize = g_pNZBQueueCoordinator->GetNZBQueueList().size();
			
	switch( m_eInputMode )
	{
		case eNormal:
			PlotLine( "(Q)uit | (P)ause/Unpause all | (Enter) Edit Queue", m_iNrRows-1, NCURSES_COLORPAIR_KEYBAR );
			wmove( m_pWindow, m_iNrRows, m_iNrColumns );
			break;
		case eEditQueue:
			if( m_iSelectedQueueEntry == 0 )
			{
				PlotLine( "(Q)uit | (D)elete", m_iNrRows-1, NCURSES_COLORPAIR_KEYBAR );
			}
			else if( m_iSelectedQueueEntry > 0 && iQueueSize > 2 && m_iSelectedQueueEntry == iQueueSize-1 )
			{
				PlotLine( "(Q)uit | (D)elete | (PgDwn) Entry up queue", m_iNrRows-1, NCURSES_COLORPAIR_KEYBAR );
			}
			else if( iQueueSize > 2 && m_iSelectedQueueEntry == 1 )
			{
				PlotLine( "(Q)uit | (D)elete | (PgUp) Entry down queue", m_iNrRows-1, NCURSES_COLORPAIR_KEYBAR );
			}
			else
			{
				PlotLine( "(Q)uit | (D)elete | (PgUp) Entry down queue | (PgDwn) Entry up queue", m_iNrRows-1, NCURSES_COLORPAIR_KEYBAR );
			}
			
			wmove( m_pWindow, m_iNrRows, m_iNrColumns );
			break;
		case eDownloadRate:
			char szString[128];
			sprintf( szString, "Download rate: %i", m_iInputValue );
			PlotLine( szString, m_iNrRows-1, NCURSES_COLORPAIR_KEYBAR );
			wmove( m_pWindow, m_iNrRows, 20 );
			break;
	}
}

void NCursesFrontend::PrintQueue()
{
	char szBuffer[512];
	
	int iQueueMaxSize = (int)((float)m_iNrRows * fNZBQueueScreenPercentage);
		
	if( g_pNZBQueueCoordinator->GetNZBQueueList().size() == 0 )
	{
		PlotLine( "Ready to receive nzb-job", m_iLineNr++, NCURSES_COLORPAIR_TEXT );
	}
	else
	{
		std::deque<NZBQueue*>::iterator it1;

		uint iCount = 0;
		
		for( it1 = g_pNZBQueueCoordinator->GetNZBQueueList().begin(); it1 != g_pNZBQueueCoordinator->GetNZBQueueList().end(); it1++ )
		{
			NZBQueue* pQueue = *it1;

			if( m_eInputMode && iCount == m_iSelectedQueueEntry )
			{
				NCOLORON( NCURSES_COLORPAIR_TEXTHIGHL );
			}
			else
			{
				NCOLORON( NCURSES_COLORPAIR_TEXT );
			}

			const char* szPause = pQueue->GetPause() ? "(paused)" : "";
			const char* szCanceled = pQueue->GetCancelDownload() ? "(canceled)" : "";

			sprintf( szBuffer, "%s (%i/%i) %s%s", basename( pQueue->GetFileName() )
								, pQueue->GetNumberOfJobs()
								, pQueue->GetOriginalNumberOfJobs()
								, szPause
								, szCanceled );

			int iLength = strlen( szBuffer );
			int iRemain = m_iNrColumns - iLength;
			while( iRemain >=0 ) szBuffer[ iLength+iRemain-- ] = PADCHAR;
			szBuffer[ m_iNrColumns-1 ] = 0;
			
			mvprintw( m_iLineNr++, 0, "%s", szBuffer );
			
			if( m_eInputMode == eEditQueue && iCount == m_iSelectedQueueEntry )
			{
				NCOLOROFF( NCURSES_COLORPAIR_TEXTHIGHL );
			}
			else
			{
				NCOLOROFF( NCURSES_COLORPAIR_TEXT );
			}
			
			if( m_iLineNr >= iQueueMaxSize )
			{
				break;
			}
			iCount++;
		}
	}
}

void NCursesFrontend::ClearRow( int iRow )
{
	char szBuffer[512];
	int iColumn = 0;
	while( iColumn < m_iNrColumns ) szBuffer[ iColumn++ ] = PADCHAR;
	szBuffer[ m_iNrColumns-1 ] = 0;
	mvprintw( iRow, 0, "%s", szBuffer );
}

void NCursesFrontend::UpdateInput()
{
	int iKey = getch();
	
	// Normal & Edit Queue mode
	if( m_eInputMode == eNormal || m_eInputMode == eEditQueue )
	{
		switch( iKey )
		{
			// Key 'q' for quit
			case 'q':
				kill( GlobalControl::GetInstance()->GetProgramPID(), SIGINT );  // send a kill signal
				break;
			// Key 'p' for pause
			case 'p':
				if( m_eInputMode == eNormal )
				{
					GlobalControl::GetInstance()->TogglePause();
					if( GlobalControl::GetInstance()->GetPause() )
					{
						FrontendInfo( "Pausing download" );
					}
					else
					{
						FrontendInfo( "Unpausing download" );
					}
				}
				// Not supported (yet)
/*				else if( m_eInputMode == eEditQueue )
				{
					g_pNZBQueueCoordinator->PauseUnpauseNZBQueueEntry( m_iSelectedQueueEntry );
				}
*/
				break;
			// Delete entry
			case 'd':
				if( m_eInputMode == eEditQueue )
				{
				
					if( g_pNZBQueueCoordinator->EditQueueDeleteEntry( m_iSelectedQueueEntry ) )
					{
						uint iSize = g_pNZBQueueCoordinator->GetNZBQueueList().size();
						if( iSize == 0 )
						{
							m_iSelectedQueueEntry = 0;
							m_eInputMode = eNormal;
						}
						if( m_iSelectedQueueEntry >= iSize )
						{
							m_iSelectedQueueEntry = iSize-1;
						}
					}
				}
				break;
			// Return
			case 10:
				if( g_pNZBQueueCoordinator->GetNZBQueueList().size() > 0 )
				{
					if( m_eInputMode == eNormal )
					{
						m_eInputMode = eEditQueue;
					}
					else if( m_eInputMode == eEditQueue )
					{
						m_eInputMode = eNormal;
					}
					PrintKeyInputBar();
				}
				break;
			//
			case KEY_DOWN:
				if( m_iSelectedQueueEntry < g_pNZBQueueCoordinator->GetNZBQueueList().size()-1 )
				{
					m_iSelectedQueueEntry++;
				}
				
				PrintKeyInputBar();
				break;
			//
			case KEY_UP:
				if( m_iSelectedQueueEntry > 0 )
				{
					m_iSelectedQueueEntry--;
				}
				
				PrintKeyInputBar();
				break;
			case KEY_PPAGE:
				if( m_eInputMode == eEditQueue )
				{
					g_pNZBQueueCoordinator->EditQueueMoveEntryUp( m_iSelectedQueueEntry );
				}
				break;
			case KEY_NPAGE:
				if( m_eInputMode == eEditQueue )
				{
					g_pNZBQueueCoordinator->EditQueueMoveEntryDown( m_iSelectedQueueEntry );
				}
				break;
			// Download rate
			case 'r':
				m_eInputMode = eDownloadRate;
				m_iInputNumberIndex = 0;
				m_iInputValue = 0;
				PrintKeyInputBar();
				break;
		}
	}
	// Edit download rate input mode
	else if( m_eInputMode == eDownloadRate )
	{
		// Numbers
		if( m_iInputNumberIndex<4 && iKey >= '0' && iKey <= '9' )
		{
			m_iInputValue = (m_iInputValue*10) + (iKey-'0');
			PrintKeyInputBar();
			m_iInputNumberIndex++;
		}
		// Enter
		else if( iKey == 10 )
		{
			GlobalControl::GetInstance()->SetDownloadRate( m_iInputValue );
			m_eInputMode = eNormal;
			PrintKeyInputBar();
			return;
		}
		// Backspace
		else if( m_iInputNumberIndex>0 && iKey == 127 )
		{
			int iRemain = m_iInputValue%10;

			m_iInputValue = (m_iInputValue-iRemain)/10;
			m_iInputNumberIndex--;
			PrintKeyInputBar();
		}

		// Print the cursor
		NCOLORON( NCURSES_COLORPAIR_CURSOR );
		attron( A_BLINK );
		mvprintw( m_iNrRows-1, 15+m_iInputNumberIndex, " " );
		attroff( A_BLINK );
		NCOLOROFF( NCURSES_COLORPAIR_CURSOR );
		move( m_iNrRows, m_iNrColumns );
	}
}
