/*
 *  This file if part of nzbget
 *
 *  Copyright (C) 2004  Sven Henkel <sidddy@users.sourceforge.net>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Revision: 1.16 $
 * $Date: 2005/07/12 09:37:48 $
 *
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "NZBQueue.h"
#include "global.h"
#include "DownloadJob.h"
#include "Frontend.h"

extern Frontend* g_pFrontend;

NZBQueue::NZBQueue( const char* szFileName, xmlDocPtr nzbdoc ) :
m_bPause( false )
{
	m_szFileName = strdup( szFileName );

	char szBuffer[512];
	strcpy( szBuffer, szFileName );
	m_szBaseFileName = strdup( basename( szBuffer ) );
	
	m_iOriginalNumberOfJobs = 0;
    
	m_Jobs.clear();
	this->parseNZB(nzbdoc);
	
	m_iOriginalNumberOfJobs = m_Jobs.size();
	m_bCancelDownload 		= false;
	m_bExistInQueueDir		= false;
}

NZBQueue::~NZBQueue()
{
    // Cleanup
    if( m_szFileName )
    {
      free( m_szFileName );
    }

	if( m_szBaseFileName )
	{
	  free( m_szBaseFileName );
	}
    
    for (unsigned int i=0; i < m_Jobs.size(); i++)
    {
        delete m_Jobs[i];
        m_Jobs[i] = NULL;
    }
    m_Jobs.clear();
}


Job* NZBQueue::GetNextJob()
{
    Job* j = NULL;
  
    if (m_Jobs.size()>0)
    {
        j = m_Jobs[0];
        m_Jobs.erase(m_Jobs.begin());
    }
    
    return j;
}

void NZBQueue::parseSegment(xmlNodePtr node, std::vector<char*>* idlist, long long* size)
{
  if (!node)
    return;
  
  if ((node->name) && (!strcmp((char*)node->name, "segment")))
    {
      // segment-tag found

      if ((node->children) && (node->children->content))
        {
          // find segment-number and -size
          xmlAttr* attr = node->properties;
          long long lsize = -1;
          int partNumber = -1;

          while (attr)
            {
              if ((attr->name) && (!strcmp((char*)attr->name, "bytes")) && (attr->children))
                {
                  lsize = atol((char*) attr->children->content);
                }

              if ((attr->name) && (!strcmp((char*)attr->name, "number")) && (attr->children))
                {
                  partNumber = atol((char*) attr->children->content);
                }

              attr = attr->next;
            }

          if (partNumber > 0)
            {
              // make idlist big enough
              while ((int)idlist->size() < partNumber)
                idlist->push_back(NULL);

              if ((*idlist)[partNumber-1] == NULL)
                {
                  // new segment, add it!
                  char tmp[2048];
                  snprintf(tmp, 2048, "<%s>", (char*)node->children->content);
                  (*idlist)[partNumber-1] = strdup(tmp);

                  if (lsize > 0)
                    *size += lsize;
                }
            }
        }

      this->parseSegment(node->next, idlist, size);
    }
  else
    {
      this->parseSegment(node->next, idlist, size);
      this->parseSegment(node->children, idlist, size);
    }
}

void NZBQueue::parseGroup(xmlNodePtr node, std::vector<char*>* grouplist)
{
  if (!node)
    return;
  
  if ((node->name) && (!strcmp((char*)node->name, "group")))
    {
      //group-tag found
      if ((node->children) && (node->children->content))
        {
          char tmp[2048];
          snprintf(tmp, 2048, "%s", (char*)node->children->content);
          grouplist->push_back(strdup(tmp));
        }
      this->parseGroup(node->next, grouplist);
    }
  else
    {
      this->parseGroup(node->next, grouplist);
      this->parseGroup(node->children, grouplist);
    }
}

void NZBQueue::parseFile(xmlNodePtr node)
{
  if (!node)
    return;

  if ((node->name) && (!strcmp((char*)node->name, "file")))
    {
      //file-tag found
      DownloadJob* job = new DownloadJob();

      xmlAttr* attr = node->properties;

      while (attr)
        {
          if ((attr->name) && (!strcmp((char*)attr->name, "subject")) && (attr->children))
            {
              job->SetSubject((const char*)attr->children->content);
              break;
            }
          attr = attr->next;
        }

      long long size = 0;

      this->parseSegment(node->children, job->getMessageIDs(), &size);
      this->parseGroup(node->children, job->getGroups());
      job->setSize(size);
      m_Jobs.push_back(job);
      this->parseFile(node->next);
    }
  else
    {
      this->parseFile(node->children);
      this->parseFile(node->next);
    }

  debug("%s",(char*)node);
}

void NZBQueue::parseNZB(xmlDocPtr nzb)
{
  // walk through whole doc and search for segments-tags
  this->parseFile(nzb->children); 
}

long long NZBQueue::GetWaitingSize()
{
	long long result = 0;
  
	for( unsigned int i = 0; i < m_Jobs.size(); i++ )
	{
		if ((m_Jobs[i]) && (m_Jobs[i]->GetType() == JOBTYPE_DOWNLOAD))
		{
			result += ((DownloadJob*)m_Jobs[i])->getSize();
		}
      
	}
	return result;
}

void NZBQueue::PrintProgressInfo()
{
	char szBuffer[1024];
	strcpy( szBuffer, basename( m_szFileName ) );
	int iLength = strlen( szBuffer );
	sprintf( szBuffer+iLength, " (%i/%i files)", m_Jobs.size(), m_iOriginalNumberOfJobs );
	g_pFrontend->FrontendInfo( szBuffer );
}
