/*
 *  This file if part of nzbget
 *
 *  Copyright (C) 2005  Bo Cordes Petersen <placebodk@users.sourceforge.net>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Revision: 1.18 $
 * $Date: 2005/07/13 15:06:03 $
 *
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <unistd.h>
#include <sys/msg.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <dirent.h>
#include <libgen.h>

#include "global.h"
#include "NZBQueueCoordinator.h"
#include "NZBQueueMessageControl.h"

#include <libxml/parser.h>

extern Frontend* g_pFrontend;
extern Options* g_pOptions;
pthread_mutex_t g_tMutexQueueList;

#define NZBQUEUE_LIST_FILENAME	"/nzbget.queue"

extern bool CreateNZBCharBufferFromFilename( const char* szFileName, char** pBuffer, int* pBufferLength );

NZBQueueCoordinator::NZBQueueCoordinator( Options* pOptions )
{
	m_pOptions = pOptions;
	m_bServerMode = !strcmp( g_pOptions->getOption( OPTION_SERVER, NULL ), "yes" );

	if( m_bServerMode )
	{
		// Setup the queue observer
		m_pQueueMessageControl = new NZBQueueMessageControl();
		m_pQueueMessageControl->SetQueueCoordinator( this );
	}
	else
	{
		m_pQueueMessageControl = NULL;
	}
	
	m_pCoordinator		= NULL;
	m_pCurrentNZBQueue	= NULL;
	m_bSaveQueueList	= !strcmp( m_pOptions->getOption( OPTION_USEQUEUEDIR, NULL ), "yes" );

	pthread_mutex_init( &g_tMutexQueueList, NULL );
	pthread_mutex_init( &m_mutexRunning, NULL );
}

NZBQueueCoordinator::~NZBQueueCoordinator()
{
	// Cleanup
	if( m_pQueueMessageControl )
	{
		delete m_pQueueMessageControl;
		m_pQueueMessageControl = NULL;
	}
	
	if( m_pCoordinator )
	{
		delete m_pCoordinator;
		m_pCoordinator = NULL;
	}
	
	std::deque<NZBQueue*>::iterator it;
	for( it=m_NZBQueueList.begin(); it != m_NZBQueueList.end(); it++ )
	{
		NZBQueue* pNZBQueue = *it;
		delete pNZBQueue;
	}
	
	m_NZBQueueList.clear();
}

/*
 * Start the donwload
 */
void NZBQueueCoordinator::Start()
{
	// Starting a thread with the frontend
	g_pFrontend->Start();

#ifdef KDEVDEBUG
	if( LoadNZBQueueList( false ) )
	{
		LoadNZBQueueList();
	}
#else
    const char* szReloadQueue = g_pOptions->getOption( OPTION_RELOADQUEUE, "ask" );

	bool bAskReload = !strcasecmp( szReloadQueue, "ask" );
	bool bForceYes = !strcasecmp( szReloadQueue, "yes" );
	
	if( m_bServerMode && m_bSaveQueueList )
	{
		bool bOldQueueFound = LoadNZBQueueList( false );
		
		if( bAskReload && bOldQueueFound )
		{
			//printf( "nzbget has detected a previous download queue. Do you want to reload (Y/N)\n" );
			g_pFrontend->FrontendInfo( "A previous download queue was detected" );
			g_pFrontend->FrontendInfo( "Do you want to reload the queue (Y/N)" );
			g_pFrontend->FrontendInfo( "(saying no will delete the old queue)" );
						
			char c;
			do
			{
				c = getchar();
				
			} while( c != 'y' && c != 'n' );

			bool bCleanup = false;
	
			if( c=='y' )
			{
				if( LoadNZBQueueList() )
				{
					info( "Queue loaded" );
				}
				else
				{
					error( "Queue did not load correctly" );
					bCleanup = true;
				}
			}
			else
			{
				bCleanup = true;
			}

			if( bCleanup )
			{
				info( "Deleting old queue" );
				CleanupNZBQueueList();
			}
		}
		else if( bForceYes && bOldQueueFound )
		{
			LoadNZBQueueList();
		}
		else
		{
			if( bOldQueueFound )
			{
				info( "Deleting old queue" );
				CleanupNZBQueueList();
			}
		}
	}
#endif

	char szTmp[1024];

	// Start the queue observer/controller - but only in server-mode
	if( m_pQueueMessageControl )
	{
	    m_pQueueMessageControl->Start();
	}
                      
	// Run through all of the NZBQueues
	while( true )
	{
		pthread_mutex_lock( &g_tMutexQueueList );

		if( m_NZBQueueList.size() > 0 )
		{
			m_pCurrentNZBQueue = m_NZBQueueList[0];
		
			pthread_mutex_unlock( &g_tMutexQueueList );
		
			// Prinout
			sprintf( szTmp, "Downloading %s", m_pCurrentNZBQueue->GetBaseFileName() );
			info( szTmp );

			m_pCurrentNZBQueue->PrintProgressInfo();
			
			// Setting the correct name in the options
			m_pOptions->setOption( OPTION_NZBFILE, m_pCurrentNZBQueue->GetFileName() );
		
			// Create the coordinator
			m_pCoordinator = new Coordinator( m_pCurrentNZBQueue, m_pOptions );
		
			// Start the actual download of the individual files referred to in the nzb-file
			// This will terminate when the entire batch has been downloaded
			m_pCoordinator->Start();

			// Prinout
			if( !m_pCoordinator->GetDownloadCanceled() )
			{
				sprintf( szTmp, "Finished downloading %s", m_pCurrentNZBQueue->GetBaseFileName() );
			}
			else
			{
				sprintf( szTmp, "Canceled downloading %s", m_pCurrentNZBQueue->GetBaseFileName() );
			}
			
			info( szTmp );
		
			// Pop the current NZBfile - this has just been downloaded
			pthread_mutex_lock( &g_tMutexQueueList );
			m_NZBQueueList.pop_front();
			pthread_mutex_unlock( &g_tMutexQueueList );

			// Tell the frontend that the download has finished and entry has been removed from queue
			g_pFrontend->OnDownloadCompleted( m_pCurrentNZBQueue->GetBaseFileName() );
		
			// If using the queue-list/dir remove the file and resave the current queue
			if( m_bSaveQueueList )
			{
				// Remove the .nzb file from the dir
				RemoveNZBFileFromDir( m_pCurrentNZBQueue->GetBaseFileName() );

				// Save the new list
				SaveNZBQueueList();					
			}

			// Cleanup
			delete m_pCurrentNZBQueue;
			delete m_pCoordinator;
		}
		else
		{
			pthread_mutex_unlock( &g_tMutexQueueList );

			// No jobs - which mode are we downloading in?
			if( m_bServerMode )
			{
				// Sleep for a while
				usleep( 5 * 1000000 );
			}
			else
			{
				break;
			}
		}
	}
    
	// Printing the last messages
	g_pFrontend->PrintMessages();
}

NZBQueue* NZBQueueCoordinator::CreateNZBQueueFromBuffer( const char* szFileName, const char* szBuffer, int iSize, bool bCreateFileInQueueDir )
{
	char szTmp[512];
	
	// get the basename of the entire filename-thingy
	strcpy( szTmp, szFileName );
	const char* szBaseName = basename( szTmp );

	if( bCreateFileInQueueDir )
	{
		// Also add the file to the queue-dir - make sure everything went ok
		if( !SaveNZBFileToQueueDir( szBaseName, szBuffer ) )
		{
			return NULL;
		}
	}
	
	xmlDocPtr doc = xmlParseMemory( szBuffer, iSize );
	
	if( !doc )
	{
		return NULL;
	}                                  
	
	xmlCleanupParser();
	
	NZBQueue* pQueue = new NZBQueue( szBaseName, doc );
	
	xmlFreeDoc(doc);

	//
	if( m_bSaveQueueList )
	{
		pQueue->SetExistInQueueDir( true );
	}
	
	return pQueue;
}

NZBQueue* NZBQueueCoordinator::AddNZBFileToQueue( const char* szFileName, bool bUseQueueDir )
{
	int iBufferLength	= 0;
	char* szBuffer		= NULL;

	CreateNZBCharBufferFromFilename( szFileName, &szBuffer, &iBufferLength );

	// Parse the buffer and make it into a NZBQueue
	NZBQueue* pNZBQueue = CreateNZBQueueFromBuffer( szFileName, szBuffer, iBufferLength, bUseQueueDir );

	free( szBuffer );

	// Did file parse correctly?
	if( !pNZBQueue )
	{
		return NULL;
	}

	// Add NZBQueue to Qeue
	AddNZBQueueToDownloadQueue( pNZBQueue, bUseQueueDir );

	return pNZBQueue;
}

void NZBQueueCoordinator::AddNZBQueueToDownloadQueue( NZBQueue* pNZBQueue, bool bUseQueueDir )
{
	pthread_mutex_lock( &g_tMutexQueueList );
	
	m_NZBQueueList.push_back( pNZBQueue );
	
	pthread_mutex_unlock( &g_tMutexQueueList );

	if( bUseQueueDir )
	{
		SaveNZBQueueList();
	}
}

void NZBQueueCoordinator::JobStarted( ArticleJob* pJob, Connection* pConnection )
{
	debug( "NZBQueueCoordinator::JobStarted - %s", pJob->getMessageID() );
	
	SRunningDL* pRun = (SRunningDL*) malloc(sizeof(SRunningDL));
  
	pRun->m_pJob 		= pJob;
	pRun->m_pConnection	= pConnection;
	pRun->bytes 		= 0;
	pRun->lines 		= 0;
	pRun->maxlines 		= 0;
	pRun->starttime 	= ::time(NULL);

	pthread_mutex_lock(&m_mutexRunning);

	m_Running.push_back( pRun );

	pthread_mutex_unlock(&m_mutexRunning);
}
    
void NZBQueueCoordinator::JobFinished( ArticleJob* pJob, Connection* pConnection )
{
	debug( "NZBQueueCoordinator::JobFinished - %s", pJob->getMessageID() );
	
	pthread_mutex_lock(&m_mutexRunning);

	for (unsigned int i=0; i<m_Running.size(); i++)
	{
		if( (m_Running[i]) && (m_Running[i]->m_pJob == pJob) &&	(m_Running[i]->m_pConnection == pConnection) )
		{
			SRunningDL* run = m_Running[i];
			m_Running.erase(m_Running.begin()+i);
			free(run);
			break;
		}
	}

	pthread_mutex_unlock(&m_mutexRunning);
}

void NZBQueueCoordinator::JobProgress( ArticleJob* pJob, Connection* pConnection, int bytes, int lines, int maxlines)
{
//	debug( "NZBQueueCoordinator::JobProgress - %s (status %s)", pJob->getMessageID(), pConnection->GetStatusString() );
	
	pthread_mutex_lock(&m_mutexRunning);

	for (unsigned int i=0; i<m_Running.size(); i++)
	{
		if ((m_Running[i]) && (m_Running[i]->m_pJob == pJob) &&
			(m_Running[i]->m_pConnection == pConnection))
		{
			m_Running[i]->bytes = bytes;
			m_Running[i]->lines = lines;
			m_Running[i]->maxlines = maxlines;
			break;
		}
	}
  
	pthread_mutex_unlock(&m_mutexRunning);
}

float NZBQueueCoordinator::CalcCurrentDownloadSpeed()
{
	float fSpeedAllDownloads = 0;

	pthread_mutex_lock(&m_mutexRunning);

	std::vector<SRunningDL*>::iterator it;

	for( it=m_Running.begin(); it!=m_Running.end(); it++)
	{
		SRunningDL* pRunning = *it;
		if( pRunning )
		{
			float fPercent = 0.0f;

			if( pRunning->maxlines > 0 )
			{
				fPercent = ( pRunning->lines) / pRunning->maxlines;
			}

			time_t tdiff = ::time(NULL) - pRunning->starttime;

			float fSpeed = 0.0f;

			if( tdiff > 0 )
			{
				fSpeed = (pRunning->bytes / tdiff / 1024);
			}

			fSpeedAllDownloads += fSpeed;
		}
	}

	pthread_mutex_unlock(&m_mutexRunning);

	return fSpeedAllDownloads;
}

/*
 * Debug function - prints out the current list of NZBQueue files via the frontend.
 */                                                                           
void NZBQueueCoordinator::PrintNZBQueueList()
{
	g_pFrontend->FrontendQueueListEntry( "-~-~-~- Current Queue List -~-~-~-" );
	
	std::deque<NZBQueue*>::iterator it;
	
	for( it=m_NZBQueueList.begin(); it!=m_NZBQueueList.end(); it++ )
	{
		NZBQueue* pQueue = *it;
		
		g_pFrontend->FrontendQueueListEntry( pQueue->GetFileName() );
	}

	g_pFrontend->FrontendQueueListEntry( "-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-" );
}

/*                                        
 * Toggles the pause flag of the specific entry in the queue
 * NOTICE: Currently this is unsupported with the download manager
 * returns true if successful, false if operation is not possible
 */
bool NZBQueueCoordinator::EditQueuePauseUnpauseEntry( uint iEntry )
{
	if( iEntry >= m_NZBQueueList.size() )
	{
		return false;
	}
	
	m_NZBQueueList[iEntry]->TogglePause();

	return true;
}

/*                                        
 * Removes entry with index iEntry
 * returns true if successful, false if operation is not possible
 */
bool NZBQueueCoordinator::EditQueueDeleteEntry( uint iEntry )
{
	uint iSize = m_NZBQueueList.size();
	if( iEntry >= iSize )
	{
		return false;
	}

	std::deque<NZBQueue*>::iterator it;

	uint iCount = 0;	
	for( it=m_NZBQueueList.begin(); it!=m_NZBQueueList.end(); it++ )
	{
		if( iCount == iEntry )
		{
			NZBQueue* pQueue = *it;

			if( pQueue == m_pCurrentNZBQueue )
			{
				if( m_pCurrentNZBQueue->GetCancelDownload() )
				{
					return false;
				}
				else
				{
					m_pCoordinator->CancelDownload();
					return true;
				}
			}
			else
			{
				m_NZBQueueList.erase( it );
				delete pQueue;
				return true;
			}
		}
		else
		{
			iCount++;
		}
	}

	return false;
}

/*                                        
 * Moves entry identified with iEntry one step up in the queue
 * returns true if successful, false if operation is not possible
 */
bool NZBQueueCoordinator::EditQueueMoveEntryUp( uint iEntry )
{
	if( m_NZBQueueList.size() < 3 || iEntry < 2 )
	{
		return false;
	}

	std::deque<NZBQueue*>::iterator it;

	NZBQueue* pNZBQueue = m_NZBQueueList[ iEntry ];

	m_NZBQueueList.erase( m_NZBQueueList.begin() + iEntry );
	m_NZBQueueList.insert( m_NZBQueueList.begin() + (iEntry-1), pNZBQueue );

	return true;
}

/*                                        
 * Moves entry identified with iEntry one step down in the queue
 * returns true if successful, false if operation is not possible
 */
bool NZBQueueCoordinator::EditQueueMoveEntryDown( uint iEntry )
{
	if( iEntry >=  m_NZBQueueList.size()-1 || iEntry < 1 )
	{
		return false;
	}

	std::deque<NZBQueue*>::iterator it;

	NZBQueue* pNZBQueue = m_NZBQueueList[ iEntry ];

	m_NZBQueueList.erase( m_NZBQueueList.begin() + iEntry );
	m_NZBQueueList.insert( m_NZBQueueList.begin() + (iEntry+1), pNZBQueue );

	return true;
}                                                                                             

bool NZBQueueCoordinator::SaveNZBFileToQueueDir( const char* szFileName, const char* szBuffer )
{
	char szTotalPath[1024];

	const char* szServerQueueDir = g_pOptions->getOption( OPTION_SERVERQUEUEDIR, NULL );
	strcpy( szTotalPath, szServerQueueDir );
	char* szBaseName = basename( (char*)szFileName );
	strcat( szTotalPath, szBaseName  );
	
	FILE* pFile = fopen( szTotalPath, "w+" );

	if( pFile == 0 )
	{
#ifdef DEBUG
		perror( "NZBQueueCoordinator::SaveNZBFileToQueueDir - fopen1" );
#endif
		
		if( CreateQueueDir() )
		{
			pFile = fopen( szTotalPath, "w+" );
		}
	
		if( pFile <= 0 )
		{
#ifdef DEBUG
			perror( "NZBQueueCoordinator::SaveNZBFileToQueueDir - fopen2" );
#endif
			return false;
		}
	}

	fprintf( pFile, "%s", szBuffer );
	fclose( pFile );

	return true;
}

/*
 * Removes a .nzb backup stored file from the queue directory
 */
bool NZBQueueCoordinator::RemoveNZBFileFromDir( const char* szFileName )
{
	char szTotalPath[1024];

	const char* szServerQueueDir = g_pOptions->getOption( OPTION_SERVERQUEUEDIR, NULL );
	strcpy( szTotalPath, szServerQueueDir );
	char* szBaseName = basename( (char*)szFileName );
	strcat( szTotalPath, szBaseName  );

	return remove( szTotalPath );
}

/*
 * Stores the current list of NZBQueue's which are currently being downloaded
 * Returns true if successful, false if not
 */
bool NZBQueueCoordinator::SaveNZBQueueList()
{
	char szTotalPath[1024];
	
	const char* szServerQueueDir = g_pOptions->getOption( OPTION_SERVERQUEUEDIR, NULL );
	strcpy( szTotalPath, szServerQueueDir );
	strcat( szTotalPath, NZBQUEUE_LIST_FILENAME );

	// Is the queue-list empty or is there "stuff" in it?
	if( m_NZBQueueList.size() == 0 )
	{
		return remove( szTotalPath );
	}

	FILE* pFile = fopen( szTotalPath, "w+" );

	if( pFile <= 0 )
	{
#ifdef DEBUG
		perror( "NZBQueueCoordinator::SaveNZBQueueList - fopen" );
#endif
		return false;
	}

	std::deque<NZBQueue*>::iterator it;

	pthread_mutex_lock( &g_tMutexQueueList );

	for( it=m_NZBQueueList.begin(); it != m_NZBQueueList.end(); it++ )
	{
		NZBQueue* pNZBQueue = *it;

		fprintf( pFile, "%s\n", pNZBQueue->GetFileName() );
	}
	
	pthread_mutex_unlock( &g_tMutexQueueList );

	fclose( pFile );

	return true;
}

/*
 * Loads and parses a list of NZBQueue files
 * returns true if successful, false if not
 * if bParseFiles is false the function only tries to see if it can open all of the files referred to in the queue file
 */
bool NZBQueueCoordinator::LoadNZBQueueList( bool bParseFiles )
{
	char szBuffer[1024];
	char szPath[1024];
	const char* szServerQueueDir = g_pOptions->getOption( OPTION_SERVERQUEUEDIR, NULL );
	strcpy( szPath, szServerQueueDir );
	strcat( szPath, NZBQUEUE_LIST_FILENAME );
	FILE* pFile = fopen( szPath, "r" );

	if( pFile <= 0 )
	{
		return false;
	}

	// Go through all the entries in the queue.xml file
	char* szRead;
	while( ( szRead = fgets( szBuffer, 1024, pFile ) ) )
	{
		strcpy( szPath, szServerQueueDir );
		strcat( szPath, szRead );

		int iLen = strlen( szPath );

		if( szPath[iLen-1] == '\n' )
		{
			szPath[iLen-1] = 0;
		}
                                                                   
		// Load, parse the file and add it to the queue
		if( bParseFiles && !AddNZBFileToQueue( szPath, false ) )
		{
			return false;
		}
		// Else just check if the file can be opened/loaded
		else if( !bParseFiles )
		{
			FILE* pTmpFile = fopen( szPath, "r" );
			if( pTmpFile <= 0 )
			{
				return false;
			}
			else
			{
				fclose( pTmpFile );
			}
		}
	}

	return true;
}

/*
 * Creates the directory in which the received .nzb files (via TCP) is backed up
 * Returns true if successful, false if not
 */
bool NZBQueueCoordinator::CreateQueueDir()
{
	const char* szServerQueueDir = g_pOptions->getOption( OPTION_SERVERQUEUEDIR, NULL );
	
	return ( mkdir( szServerQueueDir, S_IRUSR | S_IWUSR | S_IXUSR | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH ) == 0 );
}

/*
 * Read the nzbget.queue file (contains all the current queued entries) and delete all references
 * Also deletes the nzbget.queue file afterwards
 * Returns true if successful, false if not
 */
bool NZBQueueCoordinator::CleanupNZBQueueList()
{
	// TEMP
	return false;
	
	char szBuffer[1024];
	char szPath[1024];
	char szFile[1024];
	const char* szServerQueueDir = g_pOptions->getOption( OPTION_SERVERQUEUEDIR, NULL );
	strcpy( szPath, szServerQueueDir );
	strcat( szPath, NZBQUEUE_LIST_FILENAME );
	FILE* pFile = fopen( szPath, "r" );

	if( pFile <= 0 )
	{
		return false;
	}

	// Go through all the entries in the queue.xml file
	char* szRead;
	while( ( szRead = fgets( szBuffer, 1024, pFile ) ) )
	{
		strcpy( szFile, szServerQueueDir );
		strcat( szFile, szRead );

		int iLen = strlen( szFile );

		if( szFile[iLen-1] == '\n' )
		{
			szFile[iLen-1] = 0;
		}

		if( remove( szFile ) )
		{
#ifdef DEBUG
			perror( "NZBQueueCoordinator::CleanupNZBQueueList() - remove1" );
#endif
		}	
	}

	// Remove the queue-file
	if( remove( szPath ) )
	{
#ifdef DEBUG
		perror( "NZBQueueCoordinator::CleanupNZBQueueList() - remove2" );
#endif
	}

	return true;
}
