/*
 *  This file if part of nzbget
 *
 *  Copyright (C) 2004  Sven Henkel <sidddy@users.sourceforge.net>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Revision: 1.14 $
 * $Date: 2005/07/13 15:06:03 $
 *
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <errno.h>
#include <signal.h>

#include <libxml/parser.h>

#include "MultiServerPool.h"
#include "global.h"
#include "GlobalControl.h"
#include "NZBQueue.h"
#include "Options.h"
#include "Coordinator.h"
#include "Frontend.h"
#include "ColoredFrontend.h"
#include "NCursesFrontend.h"
#include "NZBQueueCoordinator.h"
#include "NZBQueueMessageControl.h"
#include "NZBClient.h"

// Externs
extern Frontend* g_pFrontend;
extern Options* g_pOptions;
extern int errno;

// Prototypes
void SigProc( int );
void Cleanup();

NZBQueueCoordinator* g_pNZBQueueCoordinator = NULL;

/*
 * Main loop
 */
int main(int argc, char *argv[])
{
	// Init options & get the name of the .nzb file
	debug( "options parsing" );
	g_pOptions					= new Options(argc, argv);
	bool bServerMode			= !strcmp( g_pOptions->getOption( OPTION_SERVER, NULL ), "yes" );
	bool bStandaloneMode		= !strcmp( g_pOptions->getOption( OPTION_STANDALONE, NULL ), "yes" );
	const char* szDocumentName	= g_pOptions->getOption( OPTION_NZBFILE, NULL );
	GlobalControl::GetInstance()->SetProgramPID( getpid() );

	if( bServerMode )
	{
		nzblog( "nzbget servermode - starting up" );
	}

	// Install a signal handler
	//	signal( SIGINT, SigProc );
	
	//------------------------------------------------
	// Not in standalone-mode
	//------------------------------------------------
	if( !bStandaloneMode && !bServerMode )
	{
		NZBClient Client;

		if( g_pOptions->m_eNZBOperation == Options::eClientRequestList )
		{
			Client.RequestServerDownloadList();
		}
		else if( g_pOptions->m_eNZBOperation == Options::eClientRequestPause )
		{
			Client.RequestServerDownloadPauseUnpause( true );
		}
		else if( g_pOptions->m_eNZBOperation == Options::eClientRequestUnpause )
		{
			Client.RequestServerDownloadPauseUnpause( false );
		}
		else if( g_pOptions->m_eNZBOperation == Options::eClientRequestSetRate )
		{
			float fRate = atof( g_pOptions->getOption( OPTION_CLIENTSETRATE, NULL ) );

			if( fRate > 0.0f )
			{
				Client.RequestServerSetDownloadRate( fRate );
			}
		}
		else
		{
			Client.RequestServerDownloadFile( szDocumentName );
		}

		return 0;
	}


	// Create the queue coordinator
	if( !g_pNZBQueueCoordinator )
	{
		g_pNZBQueueCoordinator = new NZBQueueCoordinator( g_pOptions );
	}

	//------------------------------------------------
	// Not using server-mode
	//------------------------------------------------
	if( !bServerMode || bStandaloneMode )
	{
		NZBQueue* pQueue = g_pNZBQueueCoordinator->AddNZBFileToQueue( szDocumentName, false );

		if( !pQueue )
		{
			abort("FATAL ERROR: Parsing NZB-document %s failed!!\n\n", szDocumentName ? szDocumentName : "N/A" );
	    
			return -1;
		}
	}

	// Create the front-end
	if( !strcmp(g_pOptions->getOption(OPTION_OUTPUT, NULL), "curses") )
	{
		g_pFrontend = new NCursesFrontend();
	}
	else if(!strcmp(g_pOptions->getOption(OPTION_OUTPUT, NULL), "colored") )
	{
		g_pFrontend = new ColoredFrontend();
	}
	else
	{
		g_pFrontend = new Frontend();
	}

	g_pNZBQueueCoordinator->Start();
		
	printf("Download finished.                         \n\n");
	
	Cleanup();
	
	return 0;
}

/*
 * Signal handler for SIGINT - catch ctrl+c and cleanup IPC and other stuff
 */
void SigProc( int iSignal )
{
	Cleanup();

	signal( SIGINT, SIG_DFL ); // Reset the signal handler
	
	kill( getpid(), SIGINT );  // send a kill signal
}

void Cleanup()
{
	if( g_pFrontend )
	{
		delete g_pFrontend;
		g_pFrontend = NULL;
	}
	
	if( g_pNZBQueueCoordinator )
	{
		delete g_pNZBQueueCoordinator;
		g_pNZBQueueCoordinator = NULL;
	}
}
